from .server import PulsarDS, DEFAULT_PULSAR_STORE_ADDRESS, pulsards_url
from .client import COMMANDS_INFO, redis_to_py_pattern
from .parser import (PyRedisParser, RedisParser, redis_parser,
                     RedisError, ResponseError,
                     InvalidResponse, NoScriptError, CommandError)


__all__ = ['PulsarDS', 'DEFAULT_PULSAR_STORE_ADDRESS', 'pulsards_url',
           'COMMANDS_INFO', 'redis_to_py_pattern',
           'PyRedisParser', 'RedisParser', 'redis_parser',
           'RedisError', 'ResponseError',
           'InvalidResponse', 'NoScriptError', 'CommandError']



import uuid
import threading
from asyncio import sleep

from pulsar import is_async


class LockError(Exception):
    pass


class RedisScript(object):
    '''An executable Lua script object
    '''
    def __init__(self, script):
        self.script = script
        self.sha = None

    def __call__(self, client, keys=None, args=None):
        '''Execute the script, passing any required ``args``
        '''
        if self.sha not in client.store.loaded_scripts:
            sha = yield from client.immediate_execute('SCRIPT', 'LOAD',
                                                      self.script)
            self.sha = sha.decode('utf-8')
            client.store.loaded_scripts.add(self.sha)

        result = client.evalsha(self.sha, keys, args)
        if is_async(result):
            result = yield from result
        return result


class Lock(object):
    '''Asynchronous locking primitive for distributing computing

    A primitive lock is in one of two states, 'locked' or 'unlocked'.
    It is created in the unlocked state. It has two basic methods,
    :meth:`.acquire` and :meth:`.release. When the state is unlocked,
    :meth:`.acquire` changes the state to locked and returns immediately.

    When the state is locked, :meth:`.acquire` wait
    until a call to :meth:`.release` changes it to unlocked,
    then the :meth:`.acquire` call resets it to locked and returns.
    '''
    def __init__(self, client, name, timeout=None, blocking=0, sleep=0.2):
        self._local = threading.local()
        self._local.token = None
        self.client = client
        self.name = name
        self.timeout = timeout
        self.blocking = blocking
        self.sleep = sleep
        if self.blocking and self.sleep > self.blocking:
            raise LockError("'sleep' must be less than 'blocking'")

    @property
    def _loop(self):
        return self.client._loop

    @property
    def token(self):
        ''''Return the token that acquire the lock or None.
        '''
        return self._local.token

    def _acquire(self):
        token = uuid.uuid1().hex.encode('utf-8')
        timeout = self.timeout and int(self.timeout * 1000) or ''
        acquired = yield from self.lua_acquire(self.client,
                                               keys=[self.name],
                                               args=[token, timeout])
        if acquired:
            self._local.token = token

        return acquired

    def acquire(self):
        loop = self._loop
        start = loop.time()
        acquired = yield from self._acquire()
        while self.blocking is not None and not acquired:
            if loop.time() - start >= self.blocking:
                break
            sleep(self.sleep)
            acquired = yield from self._acquire()

        return acquired

    def release(self):
        expected_token = self.token
        if not expected_token:
            raise LockError("Cannot release an unlocked lock")
        self._local.token = None
        released = yield from self.lua_release(self.client,
                                               keys=[self.name],
                                               args=[expected_token])
        if not released:
            raise LockError("Cannot release a lock that's no longer owned")
        return True

    lua_acquire = RedisScript("""
        if redis.call('setnx', KEYS[1], ARGV[1]) == 1 then
            if ARGV[2] ~= '' then
                redis.call('pexpire', KEYS[1], ARGV[2])
            end
            return 1
        end
        return 0
    """)

    # KEYS[1] - lock name
    # ARGS[1] - token
    # return 1 if the lock was released, otherwise 0
    lua_release = RedisScript("""
        local token = redis.call('get', KEYS[1])
        if not token or token ~= ARGV[1] then
            return 0
        end
        redis.call('del', KEYS[1])
        return 1
    """)



REQUIRE_JS = ('https://cdnjs.cloudflare.com/ajax/libs/require.js/2.1.20/'
              'require.js')
BOOTSTRAP = 'https://bootswatch.com/flatly/bootstrap.min.css'

twitter_api_key = 'twitter API key of your registered twitter application'
twitter_api_secret = 'consumer secret'
twitter_access_token = 'Access token'
twitter_access_secret = 'Access token secret'

twitter_stream_filter = {'track': 'python'}



from .consts import *           # noqa
from .access import *           # noqa
from .futures import *          # noqa
from .events import *           # noqa
from .proxy import *            # noqa
from .protocols import *        # noqa
from .clients import *          # noqa
from .actor import *            # noqa
from .concurrency import *      # noqa
from . import commands          # noqa



'''
.. autoclass:: TwitterSite
   :members:
   :member-order: bysource

'''
import os
import json
from string import Template

from pulsar import task
from pulsar.apps.data import create_store
from pulsar.apps.wsgi import LazyWsgi, WsgiHandler, Router, MediaRouter
from pulsar.apps.ws import WebSocket, PubSubWS

STATIC_DIR = os.path.join(os.path.dirname(__file__), 'assets')
DESCRIPTION = ('An example of using pulsar with two actor monitors, one '
               'for a WSGI server and one for a Twitter streaming server')


class TweetsWsHandler(PubSubWS):
    '''WebSocket Handler for new tweets
    '''
    def write(self, websocket, message):
        message = message.decode('utf-8')
        super().write(websocket, message)


class TwitterSite(LazyWsgi):
    '''A :class:`.LazyWsgi` handler for displaying tweets
    '''
    def __init__(self, channel):
        self.channel = channel

    def setup(self, environ):
        '''Called once only by the WSGI server.

        It returns a :class:`.WsgiHandler` with three routes:

        * The base route served by the :meth:`home_page` method
        * The websocket route
        * A route for static files
        '''
        cfg = environ['pulsar.cfg']
        # Create the store and the pubsub handler
        self.store = create_store(cfg.data_store)
        pubsub = self.store.pubsub()
        # subscribe to channel
        self.subscribe(pubsub)
        return WsgiHandler([Router('/', get=self.home_page),
                            MediaRouter('/static', STATIC_DIR),
                            WebSocket('/message',
                                      TweetsWsHandler(pubsub, self.channel))])

    def home_page(self, request):
        '''Renders the home page
        '''
        doc = request.html_document
        self.doc_head(request)
        html = Template(open(os.path.join(STATIC_DIR, 'home.html')).read())
        scheme = 'wss' if request.is_secure else 'ws'
        ws = request.absolute_uri('/message', scheme=scheme)
        html = html.safe_substitute({'twitter': json.dumps({'url': ws})})
        doc.body.append(html)
        return doc.http_response(request)

    def doc_head(self, request, title=None, description=None):
        cfg = request.cfg
        description = description or DESCRIPTION

        head = request.html_document.head
        head.title = title or 'Pulsar Twitter Stream'
        head.add_meta(name='description', content=description)
        head.links.append('/static/favicon.ico', rel="icon",
                          type='image/x-icon')
        head.links.append(cfg.BOOTSTRAP)
        head.scripts.append(cfg.REQUIRE_JS)
        head.scripts.append('/static/tweets.js')

    @task
    def subscribe(self, pubsub):
        '''Subscribe to the channel for tweets
        '''
        yield from pubsub.subscribe(self.channel)



'''Pulsar HTTP test application::

    python manage.py

Implementation
======================

.. autoclass:: HttpBin
   :members:
   :member-order: bysource

Server Hooks
===================

This example shows how to use
:ref:`server hooks <setting-section-application-hooks>` to log each request

.. automodule:: examples.httpbin.config
   :members:

'''
import os
import sys
import string
import mimetypes
from functools import partial
from itertools import repeat, chain
from random import random
from base64 import b64encode
from asyncio import async

from pulsar import HttpRedirect, HttpException, version, JAPANESE, CHINESE
from pulsar.utils.httpurl import (Headers, ENCODE_URL_METHODS,
                                  ENCODE_BODY_METHODS)
from pulsar.utils.html import escape
from pulsar.apps import wsgi, ws
from pulsar.apps.wsgi import (route, Html, Json, HtmlDocument, GZipMiddleware,
                              String)
from pulsar.utils.structures import MultiValueDict
from pulsar.utils.system import json

METHODS = frozenset(chain((m.lower() for m in ENCODE_URL_METHODS),
                          (m.lower() for m in ENCODE_BODY_METHODS)))
pyversion = '.'.join(map(str, sys.version_info[:3]))
ASSET_DIR = os.path.join(os.path.dirname(os.path.abspath(__file__)), 'assets')
FAVICON = os.path.join(ASSET_DIR, 'favicon.ico')
characters = string.ascii_letters + string.digits


def asset(name, mode='r', chunk_size=None):
    name = os.path.join(ASSET_DIR, name)
    if os.path.isfile(name):
        content_type, encoding = mimetypes.guess_type(name)
        if chunk_size:
            def _chunks():
                with open(name, mode) as file:
                    while True:
                        data = file.read(chunk_size)
                        if not data:
                            break
                        yield data
            data = _chunks()
        else:
            with open(name, mode) as file:
                data = file.read()
        return data, content_type, encoding


class BaseRouter(wsgi.Router):
    ########################################################################
    #    INTERNALS
    def bind_server_event(self, request, event, handler):
        consumer = request.environ['pulsar.connection'].current_consumer()
        consumer.bind_event(event, handler)

    def info_data_response(self, request, **params):
        data = self.info_data(request, **params)
        return Json(data).http_response(request)

    def info_data(self, request, **params):
        headers = self.getheaders(request)
        data = {'method': request.method,
                'headers': headers,
                'pulsar': self.pulsar_info(request)}
        if request.method in ENCODE_URL_METHODS:
            data['args'] = dict(request.url_data)
        else:
            args, files = request.data_and_files()
            jfiles = MultiValueDict()
            if files:
                for name, parts in files.lists():
                    for part in parts:
                        try:
                            part = part.string()
                        except UnicodeError:
                            part = part.base64()
                        jfiles[name] = part
            if isinstance(args, MultiValueDict):
                args = dict(args)

            data.update((('args', args),
                         ('files', dict(jfiles))))
        data.update(params)
        return data

    def getheaders(self, request):
        headers = Headers(kind='client')
        for k in request.environ:
            if k.startswith('HTTP_'):
                headers[k[5:].replace('_', '-')] = request.environ[k]
        return dict(headers)

    def pulsar_info(self, request):
        return request.get('pulsar.connection').info()


class HttpBin(BaseRouter):
    '''The main :class:`.Router` for the HttpBin application
    '''
    def get(self, request):
        '''The home page of this router'''
        ul = Html('ul')
        for router in sorted(self.routes, key=lambda r: r.creation_count):
            a = router.link(escape(router.route.path))
            a.addClass(router.name)
            for method in METHODS:
                if router.getparam(method):
                    a.addClass(method)
            li = Html('li', a, ' %s' % router.getparam('title', ''))
            ul.append(li)
        title = 'Pulsar'
        html = request.html_document
        html.head.title = title
        html.head.links.append('httpbin.css')
        html.head.links.append('favicon.ico', rel="icon", type='image/x-icon')
        html.head.scripts.append('httpbin.js')
        ul = ul.render(request)
        templ, _, _ = asset('template.html')
        body = templ % (title, JAPANESE, CHINESE, version, pyversion, ul)
        html.body.append(body)
        return html.http_response(request)

    @route(title='Returns GET data')
    def get_get(self, request):
        return self.info_data_response(request)

    @route(title='Returns POST data')
    def post_post(self, request):
        return self.info_data_response(request)

    @route(title='Returns PATCH data')
    def patch_patch(self, request):
        return self.info_data_response(request)

    @route(title='Returns PUT data')
    def put_put(self, request):
        return self.info_data_response(request)

    @route(title='Returns DELETE data')
    def delete_delete(self, request):
        return self.info_data_response(request)

    @route('redirect/<int(min=1,max=10):times>', defaults={'times': 5},
           title='302 Redirect n times')
    def redirect(self, request):
        num = request.urlargs['times'] - 1
        if num:
            raise HttpRedirect('/redirect/%s' % num)
        else:
            raise HttpRedirect('/get')

    @route('getsize/<int(min=1,max=8388608):size>', defaults={'size': 150000},
           title='Returns a preset size of data (limit at 8MB)')
    def getsize(self, request):
        size = request.urlargs['size']
        data = {'size': size, 'data': 'd' * size}
        return self.info_data_response(request, **data)

    @route(title='Returns gzip encoded data')
    def gzip(self, request):
        response = self.info_data_response(request, gzipped=True)
        return GZipMiddleware(10)(request.environ, response)

    @route(title='Returns cookie data')
    def cookies(self, request):
        cookies = request.cookies
        d = dict(((c.key, c.value) for c in cookies.values()))
        return Json({'cookies': d}).http_response(request)

    @route('cookies/set/<name>/<value>', title='Sets a simple cookie',
           defaults={'name': 'package', 'value': 'pulsar'})
    def request_cookies_set(self, request):
        key = request.urlargs['name']
        value = request.urlargs['value']
        request.response.set_cookie(key, value=value)
        request.response.status_code = 302
        request.response.headers['location'] = '/cookies'
        return request.response

    @route('status/<int(min=100,max=505):status>',
           title='Returns given HTTP Status code',
           defaults={'status': 418})
    def status(self, request):
        request.response.content_type = 'text/html'
        raise HttpException(status=request.urlargs['status'])

    @route(title='Returns response headers')
    def response_headers(self, request):
        class Gen:
            headers = None

            def __call__(self, server, **kw):
                self.headers = server.headers

            def generate(self):
                # yield a byte so that headers are sent
                yield b''
                # we must have the headers now
                yield json.dumps(dict(self.headers))
        gen = Gen()
        self.bind_server_event(request, 'on_headers', gen)
        request.response.content = gen.generate()
        request.response.content_type = 'application/json'
        return request.response

    @route('basic-auth/<username>/<password>',
           title='Challenges HTTPBasic Auth',
           defaults={'username': 'username', 'password': 'password'})
    def challenge_auth(self, request):
        auth = request.get('http.authorization')
        if auth and auth.authenticated(request.environ, **request.urlargs):
            return Json({'authenticated': True,
                         'username': auth.username}).http_response(request)
        raise wsgi.HttpAuthenticate('basic')

    @route('digest-auth/<username>/<password>/<qop>',
           title='Challenges HTTP Digest Auth',
           defaults={'username': 'username',
                     'password': 'password',
                     'qop': 'auth'})
    def challenge_digest_auth(self, request):
        auth = request.get('http.authorization')
        if auth and auth.authenticated(request.environ, **request.urlargs):
            return Json({'authenticated': True,
                         'username': auth.username}).http_response(request)
        raise wsgi.HttpAuthenticate('digest', qop=[request.urlargs['qop']])

    @route('stream/<int(min=1):m>/<int(min=1):n>',
           title='Stream m chunk of data n times',
           defaults={'m': 300, 'n': 20})
    def request_stream(self, request):
        m = request.urlargs['m']
        n = request.urlargs['n']
        request.response.content_type = 'text/plain'
        request.response.content = repeat(b'a' * m, n)
        return request.response

    @route(title='A web socket graph')
    def websocket(self, request):
        data = open(os.path.join(os.path.dirname(__file__),
                                 'assets', 'websocket.html')).read()
        scheme = 'wss' if request.is_secure else 'ws'
        host = request.get('HTTP_HOST')
        data = data % {'address': '%s://%s/graph-data' % (scheme, host)}
        request.response.content_type = 'text/html'
        request.response.content = data
        return request.response

    @route(title='Live server statistics')
    def stats(self, request):
        '''Live stats for the server.

        Try sending lots of requests
        '''
        # scheme = 'wss' if request.is_secure else 'ws'
        # host = request.get('HTTP_HOST')
        # address = '%s://%s/stats' % (scheme, host)
        doc = HtmlDocument(title='Live server stats', media_path='/assets/')
        # docs.head.scripts
        return doc.http_response(request)

    @route('clip/<int(min=256,max=16777216):chunk_size>',
           defaults={'chunk_size': 4096},
           title='Show a video clip')
    def clip(self, request):
        c = request.urlargs['chunk_size']
        data, ct, encoding = asset('clip.mp4', 'rb', chunk_size=c)
        response = request.response
        response.content_type = ct
        response.encoding = encoding
        response.content = data
        return response

    @route('servername',
           title='display the server name')
    def servername(self, request):
        name = request.get('SERVER_NAME')
        return String(name, '\n').http_response(request)

    ########################################################################
    #    BENCHMARK ROUTES
    @route()
    def json(self, request):
        return Json({'message': "Hello, World!"}).http_response(request)

    @route()
    def plaintext(self, request):
        return String('Hello, World!').http_response(request)


class Upload(BaseRouter):
    response_content_types = ['multipart/form-data']

    def put(self, request):
        return async(self._async_put(request))

    def _async_put(self, request):
        headers = self.getheaders(request)
        data = {'method': request.method,
                'headers': headers,
                'pulsar': self.pulsar_info(request),
                'args': MultiValueDict(),
                'files': MultiValueDict()}
        request.cache.response_data = data
        yield from request.data_and_files(stream=partial(self.stream, request))
        data['args'] = dict(data['args'])
        data['files'] = dict(data['files'])
        return Json(data).http_response(request)

    def stream(self, request, part):
        if request.cache.current_data is not part:
            request.cache.current_data = part
            request.cache.current_data_buffer = []

        request.cache.current_data_buffer.append(part.recv())

        if part.complete():
            data_store = request.cache.response_data
            data = b''.join(request.cache.current_data_buffer)
            store = data_store['args']
            if part.is_file():
                store = data_store['files']
                try:
                    data.decode('utf-8')
                except UnicodeError:
                    data = b64encode(data)

            store[part.name] = data.decode('utf-8')


class ExpectFail(BaseRouter):

    def post(self, request):
        request.get('wsgi.input').fail()


class Graph(ws.WS):

    def on_message(self, websocket, msg):
        websocket.write(json.dumps([(i, random()) for i in range(100)]))


class Site(wsgi.LazyWsgi):

    def setup(self, environ):
        router = HttpBin('/')
        return wsgi.WsgiHandler([ExpectFail('expect'),
                                 Upload('upload'),
                                 wsgi.wait_for_body_middleware,
                                 wsgi.clean_path_middleware,
                                 wsgi.authorization_middleware,
                                 wsgi.MediaRouter('media', ASSET_DIR,
                                                  show_indexes=True),
                                 ws.WebSocket('/graph-data', Graph()),
                                 router],
                                async=True)


def server(description=None, **kwargs):
    description = description or 'Pulsar HttpBin'
    return wsgi.WSGIServer(Site(), description=description, **kwargs)


if __name__ == '__main__':  # pragma    nocover
    server().start()



'''
.. autoclass:: Twitter
   :members:
   :member-order: bysource

'''
import pulsar
from pulsar import task
from pulsar.utils.system import json
from pulsar.apps.data import create_store
from pulsar.apps.http import HttpClient, OAuth1


class Twitter(pulsar.Application):
    '''A pulsar :class:`.Application` for connecting and streaming
    tweets from `twitter streaming api`_.
    This application requires the following parameters
    to be specified in your ``config.py`` file:

    * ``twitter_api_key`` the Consumer key of your application
    * ``twitter_api_secret``, the Consumer secret
    * ``twitter_access_token``, the application Access token
    * ``twitter_access_secret``, the Access token secret
    * ``twitter_stream_filter``, dictionary of parameters for
      `filtering tweets`_.
    '''
    interval1 = 0
    interval2 = 0
    interval3 = 0
    public_stream = 'https://stream.twitter.com/1.1/statuses/filter.json'

    def monitor_start(self, monitor):
        '''When the monitor starts make sure only one worker at most
        is spawned
        '''
        self.cfg.set('workers', min(self.cfg.workers, 1))

    def worker_start(self, worker, exc=None):
        '''When the worker starts :meth:`connect` to twitter spreaming
        endpoint
        '''
        api_key = self.get_param('twitter_api_key')
        client_secret = self.get_param('twitter_api_secret')
        access_token = self.get_param('twitter_access_token')
        access_secret = self.get_param('twitter_access_secret')
        self._http = HttpClient(encode_multipart=False)
        oauth1 = OAuth1(api_key,
                        client_secret=client_secret,
                        resource_owner_key=access_token,
                        resource_owner_secret=access_secret)
        self._http.bind_event('pre_request', oauth1)
        self.buffer = []
        self.connect()

    @task
    def connect(self):
        '''Connect to twitter streaming endpoint.

        If the connection is dropped, the :meth:`reconnect` method is invoked
        according to twitter streaming connection policy_.
        '''
        filter = self.get_param('twitter_stream_filter')
        return self._http.post(self.public_stream,
                               data=filter,
                               on_headers=self.connected,
                               data_processed=self.process_data,
                               post_request=self.reconnect)

    def connected(self, response, **kw):
        '''Callback when a succesful connection is made.

        Reset reconnection intervals to 0
        '''
        if response.status_code == 200:
            self.logger.info('Successfully connected with twitter streaming')
            self.interval1 = 0
            self.interval2 = 0
            self.interval3 = 0

    def process_data(self, response, **kw):
        '''Callback passed to :class:`HttpClient` for processing
        streaming data.
        '''
        if response.status_code == 200:
            messages = []
            data = response.recv_body()
            while data:
                idx = data.find(b'\r\n')
                if idx < 0:     # incomplete data - add to buffer
                    self.buffer.append(data)
                    data = None
                else:
                    self.buffer.append(data[:idx])
                    data = data[idx+2:]
                    msg = b''.join(self.buffer)
                    self.buffer = []
                    if msg:
                        body = json.loads(msg.decode('utf-8'))
                        if 'disconnect' in body:
                            msg = body['disconnect']
                            self.logger.warning('Disconnecting (%d): %s',
                                                msg['code'], msg['reason'])
                        elif 'warning' in body:
                            message = body['warning']['message']
                            self.logger.warning(message)
                        else:
                            messages.append(body)
            if messages:
                # a list of messages is available
                if self.cfg.callable:
                    self.cfg.callable(self, messages)

    def reconnect(self, response, exc=None):
        '''Handle reconnection according to twitter streaming policy_

        .. _policy: https://dev.twitter.com/docs/streaming-apis/connecting
        '''
        loop = response._loop
        if response.status_code == 200:
            gap = 0
        elif not response.status_code:
            # This is a network error, back off lineraly 250ms up to 16s
            self.interval1 = gap = max(self.interval1+0.25, 16)
        elif response.status_code == 420:
            gap = 60 if not self.interval2 else max(2*self.interval2)
            self.interval2 = gap
        else:
            if response.status_code >= 400:
                self.logger.error('Could not connect to twitter spreaming API,'
                                  ' status code %s' % response.status_code)
            gap = 5 if not self.interval3 else max(2*self.interval3, 320)
            self.interval3 = gap

        loop.call_later(gap, self.connect)

    def get_param(self, name):
        value = self.cfg.get(name)
        if not value:
            raise pulsar.ImproperlyConfigured(
                'Please specify the "%s" parameter in your %s file' %
                (name, self.cfg.config))
        return value


class PublishTweets:
    '''Tweets processor

    This callable class is passed to the :class:`.Twitter` application
    and it is invoked every time new messages are available.

    The callable method accept two parameters, the :class:`.Twitter`
    application and a list of messages received.
    '''
    store = None

    def __init__(self, channel):
        self.channel = channel

    @task
    def __call__(self, twitter, messages):
        if not self.store:
            self.store = create_store(twitter.cfg.data_store)
            self.pubsub = self.store.pubsub()
        for message in messages:
            yield from self.pubsub.publish(self.channel, json.dumps(message))
